class CombinedIters:
  def __init__(self, i1, i2):
    self.i1 = i1
    self.i2 = i2
    self.first_iter = True

  def next(self):
    if self.first_iter:
      try:
        return self.i1.next()
      except:
        self.first_iter = False
        return self.i2.next()
    else:
      return self.i2.next()

def combine_iters(*iters):
  assert len(iters) > 0
  curr_iter = iters[0]
  for i in iters[1:]:
    curr_iter = CombinedIters(curr_iter, i)
  return curr_iter

class NullIter(object):
  def next(self):
    raise StopIteration

def maybe_iter(obj):
  if obj is None:
    return NullIter()
  else:
    return iter(obj)

class MissingArg(object):
  pass 
# placeholder object 
missing_arg = MissingArg()

class MissingArgsError(Exception):
  def __init__(self, missing_arg_names, fn_name = None, file_name = None, line_no = None):
    import names
    self.missing_arg_names = [names.original(name) for name in missing_arg_names]
    
    self.fn_name = fn_name 
    self.file_name = file_name 
    self.line_no = line_no 
    
  def __str__(self):
    if len(self.missing_arg_names) > 1:
      err_str = "Missing args %s" % ", ".join(["'%s'" % name for name in self.missing_arg_names])
    else:
      err_str = "Missing arg '%s'" % self.missing_arg_names[0]
    
    if self.fn_name is not None:
      err_str += " in call to function '%s'" % self.fn_name
      
    if self.file_name is not None:
      err_str += " in file %s" % self.file_name 
    
    if self.line_no is not None:
      err_str += " on line %d" % self.line_no
    return err_str
  
class ActualArgs(object):
  def __init__(self, positional, keywords = {}, starargs = None):
    positional = tuple(positional)
    self.positional = positional
    self.keywords = keywords
    self.starargs = starargs

  def transform(self, fn, keyword_name_fn = None, keyword_value_fn = None):
    new_pos = [fn(pos_arg) for pos_arg in self.positional]
    new_keywords = {}
    for (k,v) in self.keywords.iteritems():
      new_name = keyword_name_fn(k) if keyword_name_fn else k
      new_value = keyword_value_fn(v) if keyword_value_fn else fn(v)
      new_keywords[new_name] = new_value
    new_starargs = fn(self.starargs) if self.starargs else None
    return ActualArgs(new_pos, new_keywords, new_starargs)

  def __eq__(self, other):
    if len(self.positional) != len(other.positional) or \
       len(self.keywords) != len(other.keywords) or \
       self.starargs != other.starargs or \
       any(t1 != t2 for (t1, t2) in zip(self.positional, other.positional)):
      return False
    for (k,v) in self.keywords.iteritems():
      if k not in other.keywords:
        return False
      if other.keywords[k] != v:
        return False
    return True

  def __str__(self):
    arg_strings = []
    for p in self.positional:
      arg_strings.append(str(p))
    for (k,v) in self.keywords.items():
      arg_strings.append("%s = %s" % (k,v))
    if self.starargs:
      arg_strings.append("*" + str(self.starargs))
    return ", ".join(arg_strings)

  def __hash__(self):
    kwd_tuple = tuple(self.keywords.items())
    full_tuple = self.positional + kwd_tuple + (self.starargs,)
    return hash(full_tuple)

  def __iter__(self):
    return combine_iters(iter(self.positional),
                         self.keywords.itervalues(),
                         maybe_iter(self.starargs))

  def prepend_positional(self, more_args):
    new_pos = tuple(more_args) + self.positional
    return ActualArgs(new_pos, self.keywords, self.starargs)

class FormalArgs(object):
  def __init__(self):
    self.n_args = 0
    self.nonlocals = ()
    self.positional = []

    self.defaults = {}
    self.starargs = None

    # map visible name to local SSA name
    self.local_names = {}
    # map SSA name to visible (keyword) name
    self.visible_names = {}

    # position of each local name in the bound args list
    self.positions = {}

  def _prepend(self, local_name, visible_name = None):
    self.n_args += 1
    if visible_name:
      self.local_names[visible_name] = local_name
      self.visible_names[local_name] = visible_name
    self.arg_slots = [local_name] + self.arg_slots
    for k in self.positions:
      self.positions[k] += 1
    self.positions[local_name] = 0
    self.positional = [local_name] + self.positional

  def add_positional(self, local_name, visible_name = None):
    self.n_args += 1
    if visible_name:
      self.local_names[visible_name] = local_name
      self.visible_names[local_name] = visible_name
    self.positions[local_name] = len(self.positions)
    self.positional.append(local_name)

  def prepend_nonlocal_args(self, localized_names):
    n_old_nonlocals = len(self.nonlocals)
    n_new_nonlocals = len(localized_names)
    total_nonlocals = n_old_nonlocals + n_new_nonlocals
    self.n_args += n_new_nonlocals
    self.nonlocals = self.nonlocals + tuple(localized_names)

    for (k,p) in self.positions.items():
      if p > n_old_nonlocals:
        self.positions[k] = p + total_nonlocals
    for (i, k) in enumerate(localized_names):
      self.positions[k] = n_old_nonlocals + i

  def __str__(self):
    strs = []
    for local_name in self.positional:
      if local_name in self.visible_names:
        s = "%s{%s}" % (local_name, self.visible_names[local_name])
      else:
        s = local_name
      if local_name in self.defaults:
        s += " = %s" % (self.defaults[local_name],)
      strs.append(s)
    if self.starargs:
      strs.append("*" + str(self.starargs))
    if self.nonlocals:
      strs.append("nonlocals = (%s)" % ", ".join(self.nonlocals))
    return ", ".join(strs)

  def __repr__(self):
    return "Args(positional = %s, defaults=%s, starargs = %s, nonlocal = %s)"% \
           (map(repr, self.positional),
            map(repr, self.defaults.items()),
            self.nonlocals,
            self.starargs)

  def bind(self, actuals,
           keyword_fn = None,
           tuple_elts_fn = iter,
           starargs_fn = tuple):
    """
    Like combine_with_actuals but returns a dictionary
    """

    env = {}
    values, extra = self.linearize_values(actuals, keyword_fn, tuple_elts_fn)

    for (k,v) in zip(self.nonlocals + tuple(self.positional), values):
      env[k] = v

    if self.starargs:
      env[self.starargs] = starargs_fn(extra)
    else:
      assert len(extra) == 0, "Too many args: %s" % (extra,)
    return env

  def linearize_values(self, actuals, keyword_fn = None, tuple_elts_fn = iter):
    if isinstance(actuals, (list, tuple)):
      actuals = ActualArgs(actuals)

    positional_values = actuals.positional

    if actuals.starargs:
      starargs_elts = tuple(tuple_elts_fn(actuals.starargs))
      positional_values = positional_values + starargs_elts

    keyword_values = actuals.keywords

    n = self.n_args
    result = [None] * n
    bound = [False] * n

    def assign(i, v):
      result[i] = v
      assert not bound[i], "%s appears twice in arguments" % self.positional[i]
      bound[i] = True

    if len(positional_values) > n:
      extra = list(positional_values[n:])
      positional_values = positional_values[:n]
    else:
      extra = []

    for (i,p) in enumerate(positional_values):
      assign(i, p)

    for (k,v) in keyword_values.iteritems():
      assert k in self.local_names, \
          "Unknown keyword: %s" % k
      local_name = self.local_names[k]
      assign(self.positions[local_name], v)

    for  (local_name, v) in self.defaults.iteritems():
      i = self.positions[local_name]
      if not bound[i]:
        assign(i, keyword_fn(local_name, v) if keyword_fn else v)
    arg_slots = self.nonlocals + tuple(self.positional)
    missing_args = [arg_slots[i] for i in xrange(n) if not bound[i]]
    if len(missing_args) > 0:
      raise MissingArgsError(missing_args)
    return result, extra
  

  def linearize_without_defaults(self, actuals, tuple_elts_fn = tuple):
    linear_args, extra = \
        self.linearize_values(actuals, tuple_elts_fn = tuple_elts_fn,
                              keyword_fn = lambda k, v: missing_arg)
    return [x for x in (linear_args + extra) if x is not missing_arg]

  def transform(self, rename_fn = lambda x: x, keyword_value_fn = None):
    args = FormalArgs()

    args.prepend_nonlocal_args(map(rename_fn, self.nonlocals))

    for old_local_name in self.positional:
      new_local_name = rename_fn(old_local_name)
      visible_name = self.visible_names.get(old_local_name)
      args.add_positional(new_local_name, visible_name)
      if old_local_name in self.defaults:
        v = self.defaults[old_local_name]
        if keyword_value_fn:
          v = keyword_value_fn(new_local_name, v)
        args.defaults[new_local_name] = v
    args.starargs = rename_fn(self.starargs) if self.starargs else None
    return args



import config

from clone_function import CloneFunction
from syntax import TypedFn

def apply_transforms(fn, transforms, cleanup = []):
  for T in transforms:
    t = T() if type(T) == type else T
    fn = t.apply(fn)
    assert fn is not None, "%s transformed fn into None" % T
    fn = apply_transforms(fn, cleanup, [])
  return fn

class Phase(object):
  def __init__(self,
               transforms,
               depends_on = [],
               copy = False,
               cleanup = [],
               config_param = None,
               run_if = None,
               rename = False,
               post_apply = None,
               memoize = True,
               name = None):
    self.cache = {}
    if not isinstance(transforms, (tuple, list)):
      transforms = [transforms]
    self.transforms = transforms

    if not depends_on:
      depends_on = []
    if not isinstance(depends_on, (tuple, list)):
      depends_on = [depends_on]
    self.depends_on = depends_on

    self.copy = copy

    if not cleanup:
      cleanup = []
    if not isinstance(cleanup, (list,tuple)):
      cleanup = [cleanup]
    self.cleanup = cleanup

    self.config_param = config_param
    self.run_if = run_if
    self.rename = rename
    self.post_apply = post_apply
    self.memoize = memoize
    self.name = name

  def __str__(self):
    if self.name:
      return self.name
    else:
      names = []
      for t in self.transforms:
        if type(t) == type:
          names.append(t.__name__)
        else:
          names.append(str(t))
      if len(names) == 1:
        return names[0]
      else:
        return "Phase(%s)" % (",".join(names))

  def __repr__(self):
    return str(self)

  def __eq__(self, other):
    return self is other

  def __call__(self, fn, run_dependencies = True, ignore_config = False):
    return self.apply(fn, run_dependencies, ignore_config)

  def apply(self, fn, run_dependencies = True, ignore_config = False):
    if self.config_param is not None and not ignore_config and \
       getattr(config, self.config_param) == False:
      return fn

    original_key = fn.name, fn.copied_by
    if original_key in self.cache:
      return self.cache[original_key]

    if self.depends_on and run_dependencies:
      fn = apply_transforms(fn, self.depends_on)

    if self.copy:
      fn = CloneFunction(self.rename).apply(fn)
      fn.copied_by = self

    if (self.run_if is None) or self.run_if(fn):
      fn = apply_transforms(fn, self.transforms, cleanup = self.cleanup)
      fn.version += 1

    if self.post_apply:
      new_fn = self.post_apply(fn)
      if new_fn.__class__ is TypedFn:
        fn = new_fn

    if self.memoize:
      self.cache[original_key] = fn
      new_key = fn.name, fn.copied_by
      self.cache[new_key] = fn

    return fn



import numpy as np 
import time 

import testing_helpers
import parakeet 

parakeet.config.print_x86 = True
parakeet.config.print_untyped_function = True 
parakeet.config.print_lowered_function = True
parakeet.config.print_optimized_llvm = True
parakeet.config.print_specialized_function = True
parakeet.config.opt_loop_unrolling = False 
parakeet.config.stride_specialization = True

def count_thresh_orig(values, thresh):
  n = 0
  for elt in values:
    n += elt < thresh
  return n

count_thresh = parakeet.jit(count_thresh_orig)

def np_thresh(values, thresh):
  return np.sum(values < thresh)

def par_thresh(values, thresh):
  return parakeet.sum(values < thresh)

"""
def gini_orig(values, thresh):
  n_left = 0
  n_right = 0
  for elt in values:
    n_left = n_left + (elt < thresh)
    n_right = n_right + (not (elt<thresh))
  total = 1.0 * n_left + n_right
  return 1.0 - (n_left/total) ** 2 - (n_right/total) ** 2

gini = parakeet.jit(gini_orig)
"""
def test_count_thresh():
  import ast 
  import inspect
  p = ast.parse(inspect.getsource(count_thresh_orig))
  print "AST"
  print ast.dump(p)
  print "Bytecode"
  import dis
  dis.dis(count_thresh_orig)
  v = np.array([1.2, 1.4, 5.0, 2, 3])
  parakeet_result = count_thresh(v, 2.0)
  python_result = count_thresh_orig(v,2.0)
  assert parakeet_result == python_result, \
    "Parakeet %s != Python %s" % (parakeet_result, python_result)
    
  v = np.random.randn(10**4)
  py_start = time.time()
  count_thresh_orig(v, 2.0)
  py_time = time.time() - py_start 

  np_start = time.time()
  np_thresh(v, 2.0)
  np_time = time.time() - np_start 

  par_start = time.time()
  count_thresh(v, 2.0)
  par_time = time.time() - par_start 

  print "Python time: %.5f" % py_time 
  print "NumPy time: %.5f" % np_time 
  print "Parakeet time: %.5f" % par_time 

if __name__ == '__main__':
  testing_helpers.run_local_tests()




from  node import Node


class Stmt(Node):
  pass

def block_to_str(stmts):
  body_str = '\n'
  body_str += '\n'.join([str(stmt) for stmt in stmts])

  return body_str.replace('\n', '\n    ')

def phi_nodes_to_str(phi_nodes):
  parts = ["%s <- phi(%s, %s)" %
           (var, left, right) for (var, (left, right)) in phi_nodes.items()]
  whole = "\n" + "\n".join(parts)
  # add tabs
  return whole.replace("\n", "\n    ")

class Assign(Stmt):
  _members = ['lhs', 'rhs']

  def __str__(self):
    if hasattr(self.lhs, 'type') and self.lhs.type:
      return "%s : %s = %s" % (self.lhs, self.lhs.type, self.rhs)
    else:
      return "%s = %s" % (self.lhs, self.rhs)

class ExprStmt(Stmt):
  """Run an expression without binding any new variables"""

  _members = ['value']

  def __str__(self):
    assert self.value is not None
    return "ExprStmt(%s)" % self.value

class Comment(Stmt):
  _members = ['text']

  def __str__(self):
    s = "#"
    for (i, c) in enumerate(self.text):
      if i % 78 == 0:
        s += "\n# "
      s += c
    s += "\n#"
    return s


class Return(Stmt):
  _members = ['value']

  def __str__(self):
    return "Return %s" % self.value

class If(Stmt):
  _members = ['cond', 'true', 'false', 'merge']

  def __str__(self):
    s = "if %s:" % self.cond
    if (len(self.true) + len(self.false)) > 0:
      s += "%s\n" % block_to_str(self.true)
    else:
      s += "\n"
    if len(self.false) > 0:
      s += "else:%s\n" % block_to_str(self.false)
    if len(self.merge) > 0:
      s += "(merge-if)%s" % phi_nodes_to_str(self.merge)
    return s

  def __repr__(self):
    return str(self)

class While(Stmt):
  """
  A loop consists of a header, which runs before each iteration, a condition for
  continuing, the body of the loop, and optionally (if we're in SSA) merge nodes
  for incoming and outgoing variables of the form
  [(new_var1, (old_var1,old_var2)]
  """

  _members = ['cond', 'body', 'merge']

  def __repr__(self):
    s = "while %s:\n  "  % self.cond
    if len(self.merge) > 0:
      s += "(header)%s\n  " % phi_nodes_to_str(self.merge)
    if len(self.body) > 0:
      s +=  "(body)%s" % block_to_str(self.body)
    return s

  def __str__(self):
    return repr(self)

class ForLoop(Stmt):
  """
  Having only one loop construct started to become cumbersome, especially now
  that we're playing with loop optimizations.

  So, here we have the stately and ancient for loop.  All hail its glory.
  """

  _members = ['var', 'start', 'stop', 'step', 'body', 'merge']
  
    
  def __str__(self):
    s = "for %s in range(%s, %s, %s):" % \
      (self.var,
       self.start.short_str(),
       self.stop.short_str(),
       self.step.short_str())

    if self.merge and len(self.merge) > 0:
      s += "\n  (header)%s\n  (body)" % phi_nodes_to_str(self.merge)
    s += block_to_str(self.body)
    return s
  


import numpy as np

import parakeet
from parakeet import jit 
import testing_helpers

def init(W, E = None):
  C = W.copy()
  if not (E is None):
    C[~E] = np.inf
  for i in xrange(C.shape[0]):
    C[i,i] = 0
  return C

def loop_dists(W, E = None):
  C = init(W,E)
  m,n = C.shape
  assert m == n
  for k in range(n):
    for i in range(n):
      for j in range(n):
        C[i,j] = min(C[i,j], C[i,k] + C[k,j])
  return C


def outer_dists(W,E = None):
  C = init(W,E)
  m,n = C.shape
  assert m == n
  vertices = np.arange(n)
  for k in vertices:
    from_k = C[:, k]
    to_k = C[k, :]
    A = np.add.outer(from_k, to_k)
    mask = (A<=C)
    C = mask * A + (1-mask) * C
    A = np.add.outer(from_k, to_k)
    mask = (A<=C)
    C = mask * A + ~mask * C
  return C


def np_dists(W, E = None):
  C = init(W,E)
  n = C.shape[0]
  for k in range(n):
    C = np.minimum(C, np.tile(C[:,k].reshape(-1,1),(1,n)) + np.tile(C[k,:],(n,1)))
  return C

def parakeet_dists(W, E = None):
  C = init(W,E)
  m,n = C.shape
  assert m == n
  vertices = np.arange(n)
  def min_row(c_row, a_row):
    return map(min, c_row, a_row)
  
  for k in vertices:
    from_k = C[:, k]
    to_k = C[k, :]
    A = parakeet.allpairs(parakeet.add, from_k, to_k)
    C = parakeet.each(min_row, C, A)
  return C

def test_dists():
  n = 100
  W = (np.random.randn(n,n)*2)**2
  C1 = outer_dists(W)
  C2 = np_dists(W)
  assert np.all(np.ravel(C1 == C2)), \
      "C2 wrong: Expected %s but got %s" % (C1, C2)
      
  C3 = parakeet_dists(W)
  assert np.all(np.ravel(C1 == C3)), \
      "Parakeet wrong: Expected %s but got %s" % (C1, C3)
"""
def test_dist_perf():
  n = 21
  W = (np.random.randn(n,n)*2)**2
  
  testing_helpers.timed_test(jit(loop_dists), [W], np_dists, [W], 
                             min_speedup = 0.1)
"""

if __name__ == '__main__':
  testing_helpers.run_local_tests()



import names 
import syntax 
from syntax import TypedFn, Var, Const, Attribute, Index, PrimCall
from syntax import If, Assign, While, ExprStmt, Return, ForLoop 
from syntax import  Slice, Struct
from syntax import Tuple, TupleProj, Cast, Alloc     
from transform import Transform 

class CloneFunction(Transform):
  """
  Copy all the objects in the AST of a function
  """
  def __init__(self, rename = False, recursive = False):
    Transform.__init__(self)
    self.rename = rename 
    self.recursive = recursive
     
  def transform_Var(self, expr):
    return Var(expr.name, type = expr.type)
  
  def transform_expr(self, expr):
    c = expr.__class__
    if c is Var:
      return self.transform_Var(expr)  
    elif c is Const:
      return Const(expr.value, type = expr.type)
    elif c is Tuple:
      new_elts = tuple(self.transform_expr(elt) for elt in expr.elts)
      return Tuple(elts = new_elts, type = expr.type)
    elif c is Attribute:
      value = self.transform_expr(expr.value)
      return Attribute(value, expr.name, type = expr.type)
    elif c is Index:
      value = self.transform_expr(expr.value)
      index = self.transform_expr(expr.index)
      return Index(value, index, type = expr.type)
    elif c is PrimCall:
      args = tuple(self.transform_expr(elt) for elt in expr.args)
      return PrimCall(expr.prim, args, type = expr.type)
    elif c is TypedFn:
      if self.recursive:
        cloner = CloneFunction(rename = self.rename, recursive = True)
        return cloner.apply(expr)
      else:
        return expr 
    elif c is Slice: 
      start = self.transform_if_expr(expr.start)
      stop = self.transform_if_expr(expr.stop)
      step = self.transform_if_expr(expr.step)
      return Slice(start, stop, step, type = expr.type)
    elif c is Struct: 
      new_args = self.transform_expr_list(expr.args)
      return Struct(args = new_args, type = expr.type)
    elif c is TupleProj: 
      new_tuple = self.transform_expr(expr.tuple)
      return TupleProj(new_tuple, expr.index, type = expr.type)
    elif c is Cast:
      return Cast(self.transform_expr(expr.value), expr.type) 
    elif c is Alloc:
      return Alloc(count = self.transform_expr(expr.count), 
                   elt_type = expr.elt_type, 
                   type = expr.type)
    else:
      args = {}
      for member_name in expr.members():
        old_value = getattr(expr, member_name)
        new_value = self.transform_if_expr(old_value)
        args[member_name] = new_value
      return expr.__class__(**args)
  
  def transform_Assign(self, stmt):
    new_lhs = self.transform_expr(stmt.lhs)
    new_rhs = self.transform_expr(stmt.rhs)
    return Assign(new_lhs, new_rhs)

  def transform_ExprStmt(self, stmt):
    return ExprStmt(self.transform_expr(stmt.value))

  def transform_Return(self, stmt):
    res = Return(self.transform_expr(stmt.value))
    return res 
  
  def transform_If(self, stmt):
    new_true = self.transform_block(stmt.true)
    new_false = self.transform_block(stmt.false)
    new_merge = self.transform_merge(stmt.merge)
    new_cond = self.transform_expr(stmt.cond)
    return If(new_cond, new_true, new_false, new_merge)
  
  def transform_While(self, stmt):
    new_body = self.transform_block(stmt.body)
    new_merge = self.transform_merge(stmt.merge)
    new_cond = self.transform_expr(stmt.cond)
    return While(new_cond, new_body, new_merge)
  
  def transform_ForLoop(self, stmt):
    new_var = self.transform_expr(stmt.var)
    new_start = self.transform_expr(stmt.start)
    new_stop = self.transform_expr(stmt.stop)
    new_step = self.transform_expr(stmt.step)
    new_body = self.transform_block(stmt.body)
    new_merge = self.transform_merge(stmt.merge)
    return ForLoop(new_var, new_start, new_stop, new_step, new_body, new_merge)  
  
  def pre_apply(self, old_fn):
    new_fundef_args = dict([(m, getattr(old_fn, m)) for m in old_fn._members])
    # create a fresh function with a distinct name and the
    # transformed body and type environment
    if self.rename: 
      new_fundef_args['name'] = names.refresh(old_fn.name)
    else:
      new_fundef_args['name'] = old_fn.name
      new_fundef_args['version'] = old_fn.next_version(old_fn.name) 
    new_fundef_args['type_env'] = old_fn.type_env.copy()
    # don't need to set a new body block since we're assuming 
    # that transform_block will at least allocate a new list 
    new_fundef = syntax.TypedFn(**new_fundef_args)
    return new_fundef 




import numpy as np 
import testing_helpers 

from testing_helpers import expect

x = np.array([1,2,3,4,5])

def identity_comprehension(x):
  return [xi for xi in x]

def test_identity_comprehension():
  expect(identity_comprehension, [x], x)


def sqr_elts(x):
  return [xi**2 for xi in x]

def test_sqr_elts():
  expect(sqr_elts, [x], x**2)
  
def outer_prod(x, y):
  return [[xi * yi for xi in x] for yi in y]


def test_outer_prod():
  x = np.array([1.0,2.0,3.0])
  y = np.array([10,20])
  res = np.array(outer_prod(x,y))
  expect(outer_prod, [x,y], res)
  
def triple_nesting(x):
  return [[[zi + yi + xi for zi in x] for yi in x] for xi in x]

def test_triple_nesting():
  x = np.array([1,2,3])
  expect(triple_nesting, [x], np.array(triple_nesting(x)))

def repeat_elts(x):
  return [x[i:i+2] for i in range(len(x)-1)]

def test_repeat_elts():
  x = np.array([0,1,2,3,4,5])
  y = np.array([[0,1], [1,2], [2,3], [3,4], [4,5]])
  expect(repeat_elts, [x], y)


if __name__ == '__main__':
  testing_helpers.run_local_tests()
  



import numpy as np
import parakeet
import testing_helpers
from testing_helpers import eq 

@parakeet.jit
def range1(n):
  return range(n)

def test_range1():
  assert eq(range1(10), np.arange(10))

@parakeet.jit
def range2(start, stop):
  return range(start, stop)

def test_range2():
  assert eq(range2(100,111), np.arange(100,111)) 

@parakeet.jit
def range3(start, stop, step):
  return range(start, stop, step)

def test_range3():
  assert eq(range3(10,20,2), np.arange(10,20,2))

def test_big_step():
  assert eq(range3(10, 20, 100), np.arange(10,20,100))

if __name__ == '__main__':
  testing_helpers.run_local_tests()



class NestedBlocks(object):
  def __init__(self):
    self._blocks = []

  def push(self, block = None):
    if block is None:
      block = []
    self._blocks.append(block)

  def pop(self):
    return self._blocks.pop()

  def top(self):
    return self._blocks[-1]
  
  def current(self):
    return self._blocks[-1]

  def append(self, stmt):
    self._blocks[-1].append(stmt)
    
  def append_to_current(self, stmt):
    self._blocks[-1].append(stmt)

  def extend_current(self, stmts):
    self._blocks[-1].extend(stmts)

  def depth(self):
    return len(self._blocks)

  def __iadd__(self, stmts):
    if not isinstance(stmts, (list, tuple)):
      stmts = [stmts]
    self.extend_current(stmts)
    return self

  def __str__(self):
    return "NestedBlocks(%s)" % self._blocks


import numpy as np 
import parakeet 
from testing_helpers import expect, run_local_tests, eq, expect_allpairs 

int_mat = np.reshape(np.arange(9), (3,3))
float_mat = np.sqrt(int_mat)
bool_mat = int_mat % 2


def dot(x,y):
  return sum(x*y)

def test_dot():
  vecs = [m[0] for m in matrices]
  expect_allpairs(dot, np.dot, vecs)

def adverb_matmult(X,Y):
  return parakeet.allpairs(dot, X, Y)

matrices = [int_mat, float_mat, bool_mat]

def transposed_np_dot(x,y):
  return np.dot(x, y.T)

def test_adverb_matmult():
  expect_allpairs(adverb_matmult, transposed_np_dot, matrices)

if __name__ == '__main__':
  run_local_tests()


import closure_type
import core_types
import syntax
import tuple_type

from array_type import ArrayT, make_array_type
from core_types import ScalarT
from syntax_helpers import const_int, const_tuple, zero
from transform import Transform


class LowerStructs(Transform):
  """The only non-scalar objects should all be created as explicit Structs"""

  def transform_TypedFn(self, expr):
    import pipeline
    return pipeline.lowering.apply(expr)

  def transform_Tuple(self, expr):
    struct_args = self.transform_expr_list(expr.elts)
    return syntax.Struct(struct_args, type = expr.type)

  def transform_Slice(self, expr):
    struct_args = self.transform_expr_list([expr.start, expr.stop, expr.step])
    return syntax.Struct(struct_args, type = expr.type)

  def transform_Assign(self, stmt):
    lhs, rhs = stmt.lhs, stmt.rhs
    if isinstance(lhs, syntax.Tuple):
      for (i, lhs_elt) in enumerate(lhs.elts):
        self.assign(lhs_elt, self.tuple_proj(rhs, i), recursive = True)
    else:
      assert isinstance(lhs, (syntax.Var, syntax.Index, syntax.Attribute)), \
          "Invalid LHS: %s" % (lhs,)
      return syntax.Assign(stmt.lhs, self.transform_expr(rhs))

  def transform_Closure(self, expr):
    _ = self.transform_expr(expr.fn)
    closure_args = self.transform_expr_list(expr.args)
    closure_id = closure_type.id_of_closure_type(expr.type)
    closure_id_node = syntax.Const(closure_id, type = core_types.Int64)
    return syntax.Struct([closure_id_node] + closure_args, type = expr.type)

  def transform_TupleProj(self, expr):
    new_tuple = self.transform_expr(expr.tuple)
    assert isinstance(expr.index, int)
    tuple_t = expr.tuple.type
    field_name, field_type  = tuple_t._fields_[expr.index]
    return syntax.Attribute(new_tuple, field_name, type = field_type)

  def transform_ClosureElt(self, expr):
    new_closure = self.transform_expr(expr.closure)
    assert isinstance(expr.index, int)
    # first field is always the closure ID, so we have to
    # index 1 higher
    field_name, field_type = new_closure.type._fields_[expr.index + 1]
    return syntax.Attribute(new_closure, field_name, type = field_type)

  def array_view(self, data, shape, strides, offset, nelts):
    """Helper function used by multiple array-related transformations"""

    data = self.assign_temp(self.transform_expr(data), "data_ptr")
    data_t = data.type
    assert isinstance(data_t, core_types.PtrT), \
        "Data field of array must be a pointer, got %s" % data_t

    shape = self.assign_temp(self.transform_expr(shape), "shape")
    shape_t = shape.type
    assert isinstance(shape_t, tuple_type.TupleT), \
        "Shape of array must be a tuple, got: %s" % shape_t

    strides = self.assign_temp(self.transform_expr(strides), "strides")
    strides_t = strides.type
    assert isinstance(strides_t, tuple_type.TupleT), \
        "Strides of array must be a tuple, got: %s" % strides_t

    rank = len(shape_t.elt_types)
    strides_rank = len(strides_t.elt_types)
    assert rank == strides_rank, \
        "Shape and strides must be of same length, but got %d and %d" % \
        (rank, strides_rank)
    array_t = make_array_type(data_t.elt_type, rank)
    return syntax.Struct([data, shape, strides, offset, nelts], type = array_t)

  def transform_ArrayView(self, expr):
    array_struct = self.array_view(expr.data, expr.shape, expr.strides,
                                   expr.offset, expr.size)
    assert expr.type == array_struct.type, \
        "Mismatch between original type %s and transformed type %s" % \
        (expr.type, array_struct.type)
    return array_struct

  def transform_Array(self, expr):
    """Array literal"""

    array_t = expr.type
    assert isinstance(array_t, ArrayT)
    elt_t = array_t.elt_type
    assert isinstance(elt_t, ScalarT)

    elts = self.transform_expr_list(expr.elts)
    n = len(elts)

    ptr_t = core_types.ptr_type(elt_t)
    alloc = syntax.Alloc(elt_t, const_int(n), type = ptr_t)
    ptr_var = self.assign_temp(alloc, "data")

    for (i, elt) in enumerate(elts):
      idx = syntax.Const(i, type = core_types.Int32)
      lhs = syntax.Index(ptr_var, idx, type = elt_t)
      self.assign(lhs, elt)

    return self.array_view(ptr_var, const_tuple(n), const_tuple(1),
                           offset = const_int(0), nelts = const_int(n))


  def transform_Range(self, expr):
    diff = self.sub(expr.stop, expr.start, "range_diff")
    nelts = self.safediv(diff, expr.step, name="nelts_raw")
    result = self.alloc_array(core_types.Int64, 
                              (nelts,), 
                              name = "range_result", 
                              explicit_struct = True)
    ptr = self.attr(result, "data", "data_ptr")
    def loop_body(i):
      v = self.add(expr.start, self.mul(i, expr.step)) 
      self.setidx(ptr, i, v)
    self.loop(zero(core_types.Int64), nelts, loop_body, return_stmt = False, while_loop = False)
    return result 
  
  def transform_ConstArray(self, expr):
    assert False, "ConstArray not implemented"
    
  def transform_Ravel(self, expr):
    return self.ravel(self.transform_expr(expr.array))


import adverb_helpers
import array_type
import ast_conversion
import closure_type
import config
import core_types 
import names
import prims
import syntax
import syntax_helpers
import tuple_type
import type_conv

from args import ActualArgs, FormalArgs, MissingArgsError
from array_type import ArrayT
from closure_type import ClosureT 
from core_types import Type, Bool, IntT, Int64,  ScalarT
from core_types import NoneType, NoneT, Unknown, UnknownT
from core_types import combine_type_list, StructT
from stride_specialization import specialize
from syntax import Fn, TypedFn, Closure, ClosureElt, Var 
from syntax_helpers import get_type, get_types, unwrap_constant
from syntax_helpers import one_i64, zero_i64, none, true, false, is_false, is_true, is_zero
from transform import Transform
from tuple_type import TupleT, make_tuple_type

class InferenceFailed(Exception):
  def __init__(self, msg):
    self.msg = msg

class VarMap:
  def __init__(self):
    self._vars = {}

  def rename(self, old_name):
    new_name = names.refresh(old_name)
    self._vars[old_name] = new_name
    return new_name

  def lookup(self, old_name):
    if old_name in self._vars:
      return self._vars[old_name]
    else:
      return self.rename(old_name)

  def __str__(self):
    return "VarMap(%s)" % self._vars

def unpack_closure(closure):
  """
  Given an object which could be either a function, a function's name, a
  closure, or a closure type:
  Return the underlying untyped function and the closure arguments
  """

  if closure.__class__ is closure_type.ClosureT:
    fn, closure_args = closure.fn, closure.arg_types
  elif closure.__class__ is Closure:
    fn = closure.fn 
    closure_args = closure.args 
  elif closure.type.__class__ is closure_type.ClosureT:
    fn, arg_types = closure.type.fn, closure.type.arg_types
    closure_args = \
        [ClosureElt(closure, i, type = arg_t)
         for (i, arg_t) in enumerate(arg_types)]
  else:
    fn = closure
    closure_args = []
    # fn = syntax.Fn.registry[fn]
  return fn, closure_args

def make_typed_closure(untyped_closure, typed_fn):
  if untyped_closure.__class__ is syntax.Fn:
    return typed_fn

  assert isinstance(untyped_closure, syntax.Expr) and \
      isinstance(untyped_closure.type, closure_type.ClosureT)
  _, closure_args = unpack_closure(untyped_closure)
  if len(closure_args) == 0:
    return typed_fn
  else:
    t = closure_type.make_closure_type(typed_fn, get_types(closure_args))
    return Closure(typed_fn, closure_args, t)

def linearize_arg_types(fn, args):

  """
  Given a function object which might be one of:
    (1) a closure type
    (2) the name of an untyped function
    (3) an untyped fn object
  and some argument types which might
    (1) a list
    (2) a tuple
    (3) an ActualArgs object
  linearize the argument types with respect to
  the untyped function's argument order and return
  both the untyped function and list of arguments
  """

  untyped_fundef, closure_args = unpack_closure(fn)
  if isinstance(args, (list, tuple)):
    args = ActualArgs(args)

  if len(closure_args) > 0:
    args = args.prepend_positional(closure_args)

  def keyword_fn(_, v):
    return type_conv.typeof(v)

  linear_args, extra = \
      untyped_fundef.args.linearize_values(args, keyword_fn = keyword_fn)
  return untyped_fundef, tuple(linear_args + extra)

def tuple_elts(tup):
  return [syntax.TupleProj(tup, i, t)
          for (i,t) in enumerate(tup.type.elt_types)]

def flatten_actual_args(args):
  if isinstance(args, (list,tuple)):
    return args
  assert isinstance(args, ActualArgs), \
      "Unexpected args: %s" % (args,)
  assert len(args.keywords) == 0
  result = list(args.positional)
  if args.starargs:
    result.extend(tuple_elts(args.starargs))
  return result

def linearize_actual_args(fn, args):
    untyped_fn, closure_args = unpack_closure(fn)
    if isinstance(args, (list, tuple)):
      args = ActualArgs(args)
    args = args.prepend_positional(closure_args)

    arg_types = args.transform(syntax_helpers.get_type)

    # Drop arguments that are assigned defaults,
    # since we're assuming those are set in the body
    # of the function
    combined_args = untyped_fn.args.linearize_without_defaults(args, tuple_elts)
    return untyped_fn, combined_args, arg_types

_invoke_type_cache = {}
def invoke_result_type(fn, arg_types):
  if isinstance(fn, syntax.TypedFn):
    assert isinstance(arg_types, (list, tuple))
    assert len(arg_types) == len(fn.input_types), \
        "Type mismatch between expected inputs %s and %s" % \
        (fn.input_types, arg_types)
    assert all(t1 == t2 for (t1,t2) in zip(arg_types, fn.input_types))
    return fn.return_type

  if isinstance(arg_types, (list, tuple)):
    arg_types = ActualArgs(arg_types)

  key = (fn, arg_types)
  if key in _invoke_type_cache:
    return _invoke_type_cache[key]

  if isinstance(fn, closure_type.ClosureT):
    closure_set = closure_type.ClosureSet(fn)
  else:
    assert isinstance(fn, closure_type.ClosureSet), \
        "Invoke expected closure, but got %s" % (fn,)
    closure_set = fn

  result_type = Unknown
  for closure_t in closure_set.closures:
    typed_fundef = specialize(closure_t, arg_types)
    result_type = result_type.combine(typed_fundef.return_type)
  _invoke_type_cache[key] = result_type
  return result_type

def identity(x):
  return x
untyped_identity_function = ast_conversion.translate_function_value(identity)


class Annotator(Transform):
  
  def __init__(self, tenv, var_map):
    Transform.__init__(self)
    self.type_env = tenv 
    self.var_map = var_map
    
  
  def transform_expr(self, expr):
    if not isinstance(expr, syntax.Expr):
      expr = ast_conversion.value_to_syntax(expr)
    
    result = Transform.transform_expr(self, expr)  
    assert result.type is not None,  \
      "Unsupported expression encountered during type inference: %s" % (expr,)
    return result 
  
  
  def transform_args(self, args, flat = False):
    if isinstance(args, (list, tuple)):
      return self.transform_expr_list(args)
    else:
      new_args = args.transform(self.transform_expr)
      if flat:
        return flatten_actual_args(new_args)
      else:
        return new_args

  def transform_keywords(self, kwds_dict):
    if kwds_dict is None:
      return {}
    else:
      result = {}
      for (k,v) in kwds_dict.iteritems():
        result[k] = self.transform_expr(v)
      return result

  def keyword_types(self, kwds):
    keyword_types = {}
    for (k,v) in kwds.iteritems():
      keyword_types[k] = get_type(v)
    return keyword_types
  
  def transform_DelayUntilTyped(self, expr):
    new_values = self.transform_expr_tuple(expr.values)
    new_syntax = expr.fn(*new_values)
    assert new_syntax.type is not None
    return new_syntax
  
  def transform_TypeValue(self, expr):
    t = expr.type_value 
    assert isinstance(t, core_types.Type), "Invalid type value %s" % (t,)
    return syntax.TypeValue(t, type=core_types.TypeValueT(t))
    
  def transform_Closure(self, expr):
    new_args = self.transform_expr_list(expr.args)
    t = closure_type.make_closure_type(expr.fn, get_types(new_args))
    return Closure(expr.fn, new_args, type = t)

  def transform_Arith(self, expr):
    untyped_fn = prims.prim_wrapper(expr)
    t = closure_type.make_closure_type(untyped_fn, ())
    return Closure(untyped_fn, (), type = t)

  def transform_Fn(self, expr):
    return expr 
    #t = closure_type.make_closure_type(expr, ())
    #return Closure(expr, [], type = t)

  def transform_Call(self, expr):
    closure = self.transform_expr(expr.fn)
    args = self.transform_args(expr.args)
    if closure.type.__class__ is core_types.TypeValueT:
      assert isinstance(args, ActualArgs)
      assert len(args.positional) == 1
      assert len(args.keywords) == 0
      assert args.starargs is None 
      return self.cast(args.positional[0], closure.type.type)
    
    untyped_fn, args, arg_types = linearize_actual_args(closure, args)

    typed_fn = specialize(untyped_fn, arg_types)
    return syntax.Call(typed_fn, args, typed_fn.return_type)

  def transform_Attribute(self, expr):
    value = self.transform_expr(expr.value)
    assert isinstance(value.type, StructT)
    result_type = value.type.field_type(expr.name)
    return syntax.Attribute(value, expr.name, type = result_type)
  
  def transform_PrimCall(self, expr):
    args = self.transform_args(expr.args)
    arg_types = get_types(args)
    
    if all(isinstance(t, ScalarT) for t in arg_types):
      upcast_types = expr.prim.expected_input_types(arg_types)
      result_type = expr.prim.result_type(upcast_types)
      return syntax.PrimCall(expr.prim, args, type = result_type)
    elif expr.prim == prims.is_:
      if arg_types[0] != arg_types[1]:
        return false
      elif arg_types[0] == NoneType:
        return true
      else:
        return syntax.PrimCall(prims.is_, args, type = Bool)
    elif all(t.rank == 0 for t in arg_types):
      # arguments should then be tuples
      assert len(arg_types) == 2
      xt, yt = arg_types
      x, y = args 
      assert isinstance(xt, TupleT), \
        "Unexpected argument types (%s,%s) for operator %s" % (xt, yt, expr.prim)
      assert isinstance(yt, TupleT), \
        "Unexepcted argument types (%s,%s) for operator %s" % (xt, yt, expr.prim)
      x1 = syntax.TupleProj(x, 0, type=xt.elt_types[0])
      x2 = syntax.TupleProj(x, 1, type=xt.elt_types[1])
      y1 = syntax.TupleProj(y, 0, type=yt.elt_types[0])
      y2 = syntax.TupleProj(y, 1, type=yt.elt_types[1]) 
      if expr.prim == prims.equal:
        first = syntax.PrimCall(prims.equal, (x1, y1), type=Bool)
        second = syntax.PrimCall(prims.equal, (x2, y2), type=Bool)
        return syntax.PrimCall(prims.logical_and, (first, second), type=Bool)
      else:
        assert False, "Unsupport tuple operation %s" % expr  
    else:
      assert all(not isinstance(t, NoneT) for t in arg_types), \
        "Invalid argument types for prim %s: %s" % (expr.prim, arg_types,)
      prim_fn = prims.prim_wrapper(expr.prim)

      max_rank = adverb_helpers.max_rank(arg_types)
      arg_names = syntax_helpers.gen_data_arg_names(len(arg_types))
      untyped_broadcast_fn = \
          adverb_helpers.nested_maps(prim_fn, max_rank, arg_names)
      typed_broadcast_fn = specialize(untyped_broadcast_fn, arg_types)
      result_t = typed_broadcast_fn.return_type
      return syntax.Call(typed_broadcast_fn, args, type = result_t)

  def transform_Index(self, expr):
    value = self.transform_expr(expr.value)
    index = self.transform_expr(expr.index)
    if isinstance(value.type, TupleT):
      assert isinstance(index.type, IntT)
      assert isinstance(index, syntax.Const)
      i = index.value
      assert isinstance(i, int)
      elt_types = value.type.elt_types
      assert i < len(elt_types), \
          "Can't get element %d of length %d tuple %s : %s" % \
          (i, len(elt_types), value, value.type)
      elt_t = value.type.elt_types[i]
      return syntax.TupleProj(value, i, type = elt_t)
    else:
      result_type = value.type.index_type(index.type)
      return syntax.Index(value, index, type = result_type)

  def transform_Array(self, expr):
    new_elts = self.transform_args(expr.elts)
    elt_types = get_types(new_elts)
    common_t = combine_type_list(elt_types)
    array_t = array_type.increase_rank(common_t, 1)
    return syntax.Array(new_elts, type = array_t)

  def transform_AllocArray(self, expr):
    elt_type = expr.elt_type
    assert isinstance(elt_type, core_types.ScalarT), \
      "Invalid array element type  %s" % (elt_type)
      
    shape = self.transform_expr(expr.shape)
    if isinstance(shape, core_types.ScalarT):
      shape = self.cast(shape, Int64)
      shape = self.tuple((shape,), "array_shape")
    assert isinstance(shape, TupleT), \
      "Invalid shape %s" % (shape,)
    rank = len(shape.elt_types)
    t = array_type.make_array_type(elt_type, rank)
    return syntax.AllocArray(shape, elt_type, type = t)
  
  def transform_Range(self, expr):
    start = self.transform_expr(expr.start) if expr.start else None
    stop = self.transform_expr(expr.stop) if expr.stop else None
    step = self.transform_expr(expr.step) if expr.step else None
    array_t = array_type.ArrayT(Int64, 1)
    return syntax.Range(start, stop, step, type = array_t)

  def transform_Slice(self, expr):
    start = self.transform_expr(expr.start)
    stop = self.transform_expr(expr.stop)
    step = self.transform_expr(expr.step)
    slice_t = array_type.make_slice_type(start.type, stop.type, step.type)
    return syntax.Slice(start, stop, step, type = slice_t)

  def transform_Var(self, expr):
    old_name = expr.name
    if old_name not in self.var_map._vars:
      raise names.NameNotFound(old_name)
    new_name = self.var_map.lookup(old_name)
    assert new_name in self.type_env, \
        "Unknown var %s (previously %s)" % (new_name, old_name)
    t = self.type_env[new_name]
    return Var(new_name, type = t)

  def transform_Tuple(self, expr):
    elts = self.transform_expr_list(expr.elts)
    return self.tuple(elts)
    #elt_types = get_types(elts)
    #t = tuple_type.make_tuple_type(elt_types)
    #return syntax.Tuple(elts, type = t)

  def transform_Const(self, expr):
    return syntax.Const(expr.value, type_conv.typeof(expr.value))
  
  def transform_ConstArrayLike(self, expr):
    typed_array_value = self.transform_expr(expr.array)
    t = typed_array_value.type
    return syntax.ConstArrayLike(typed_array_value, expr.value, type = t)

  def transform_Reshape(self, expr):
    array = self.transform_expr(expr.array)
    shape = self.transform_expr(expr.shape)
    rank = len(shape.type.elt_types)
    assert isinstance(array.type, array_type.ArrayT)
    t = array_type.make_array_type(array.elt_type, rank)
    return syntax.Reshape(array, shape, type = t)
  
  def transform_Ravel(self, expr):
    array = self.transform_expr(expr.array)
    if not isinstance(array.type, array_type.ArrayT):
      print "Warning: Can't ravel/flatten an object of type %s" % array.type 
      return array 
    t = array_type.make_array_type(array.type.elt_type, 1)
    return syntax.Ravel(array, type = t)
  
  
  def transform_Cast(self, expr):
    v = self.transform_expr(expr.value)
    return syntax.Cast(v, type = expr.type)
  
  def transform_Len(self, expr):
    v = self.transform_expr(expr.value)
    t = v.type
    if t.__class__ is ArrayT:
      shape_t = make_tuple_type([Int64] * t.rank)
      shape = syntax.Attribute(v, 'shape', type = shape_t)
      return syntax.TupleProj(shape, 0, type = Int64)
    else:
      assert t.__class__ is TupleT, \
         "Unexpected argument type for 'len': %s" % t
      return syntax.Const(len(t.elt_types), type = Int64)
 
  def transform_IndexMap(self, expr):
    shape = self.transform_expr(expr.shape)
    if not isinstance(shape.type, TupleT):
      assert isinstance(shape.type, ScalarT), "Invalid shape for IndexMap: %s" % (shape,)
      shape = self.tuple((shape,))
    closure = self.transform_expr(expr.fn)
    shape_t = shape.type
    if isinstance(shape_t, IntT):
      shape = self.cast(shape, Int64)
      n_indices = 1
    else:
      assert isinstance(shape_t, TupleT), "Expected shape to be tuple, instead got %s" % (shape_t,)
      assert all(isinstance(t, ScalarT) for t in shape_t.elt_types)
      n_indices = len(shape_t.elt_types)
      if not all(t == Int64 for t in shape_t.elt_types):
        elts = tuple(self.cast(elt, Int64) for elt in self.tuple_elts(shape))
        shape = self.tuple(elts)
    result_type, typed_fn = specialize_IndexMap(closure.type, n_indices)
    return syntax.IndexMap(shape = shape, 
                           fn = make_typed_closure(closure, typed_fn), 
                           type = result_type)
  
  def transform_IndexReduce(self, expr):
    shape = self.transform_expr(expr.shape)
    map_fn_closure = self.transform_expr(expr.fn)
    combine_closure = self.transform_expr(expr.combine)
    init = self.transform_if_expr(expr.init)
    shape_t = shape.type
    if isinstance(shape_t, IntT):
      shape = self.cast(shape, Int64)
      n_indices = 1
    else:
      assert isinstance(shape_t, TupleT)
      assert all(isinstance(t, ScalarT) for t in shape_t.elt_types)
      n_indices = len(shape_t.elt_types)
      if not all(t == Int64 for t in shape_t.elt_types):
        elts = tuple(self.cast(elt, Int64) for elt in self.tuple_elts(shape))
        shape = self.tuple(elts)
    result_type, typed_fn, typed_combine = \
      specialize_IndexReduce(map_fn_closure.type, combine_closure, n_indices, init)
    if not self.is_none(init):
      init = self.cast(init, result_type)
    return syntax.IndexReduce(shape = shape, 
                              fn = make_typed_closure(map_fn_closure, typed_fn),
                              combine = make_typed_closure(combine_closure, typed_combine),
                              init = init,  
                              type = result_type)
  
  def transform_Map(self, expr):
    closure = self.transform_expr(expr.fn)
    new_args = self.transform_args(expr.args, flat = True)
    arg_types = get_types(new_args)
    assert len(arg_types) > 0, "Map requires array arguments"
    # if all arguments are scalars just handle map as a regular function call
    if all(isinstance(t, ScalarT) for t in arg_types):
      return self.invoke(closure, new_args)
    # if any arguments are tuples then all of them should be tuples of same len
    elif any(isinstance(t, TupleT) for t in arg_types):
      assert all(isinstance(t, TupleT) for t in arg_types), \
        "Map doesn't support input types %s" % (arg_types,)
      nelts = len(arg_types[0].elt_types)
      assert all(len(t.elt_types) == nelts for t in arg_types[1:]), \
       "Tuple arguments to Map must be of same length"
      zipped_elts = []
      for i in xrange(nelts):
        zipped_elts.append([self.tuple_proj(arg,i) for arg in new_args])
      return self.tuple([self.invoke(closure, elts) for elts in zipped_elts])
    axis = self.transform_if_expr(expr.axis)
    result_type, typed_fn = specialize_Map(closure.type, arg_types)
    if axis is None or self.is_none(axis):
      assert adverb_helpers.max_rank(arg_types) == 1
      axis = syntax_helpers.zero_i64
    return syntax.Map(fn = make_typed_closure(closure, typed_fn),
                       args = new_args,
                       axis = axis,
                       type = result_type)


  def flatten_Reduce(self, map_fn, combine, x, init):
    """Turn an axis-less reduction into a IndexReduce"""
    shape = self.shape(x)
    n_indices = self.rank(x)
    # build a function from indices which picks out the data elements
    # need for the original map_fn
 
    
    outer_closure_args = self.closure_elts(map_fn)
    args_obj = FormalArgs()
    inner_closure_vars = []
    for i in xrange(len(outer_closure_args)):
      visible_name = "c%d" % i
      name = names.fresh(visible_name)
      args_obj.add_positional(name, visible_name)
      inner_closure_vars.append(Var(name))
    
    data_arg_name = names.fresh("x")
    data_arg_var = Var(data_arg_name)
    idx_arg_name = names.fresh("i")
    idx_arg_var = Var(idx_arg_name)
    
    args_obj.add_positional(data_arg_name, "x")
    args_obj.add_positional(idx_arg_name, "i")
    
    idx_expr = syntax.Index(data_arg_var, idx_arg_var)
    inner_fn = self.get_fn(map_fn)
    fn_call_expr = syntax.Call(inner_fn, tuple(inner_closure_vars)  + (idx_expr,))
    idx_fn = syntax.Fn(name = names.fresh("idx_map"),
                       args = args_obj, 
                       body =  [syntax.Return(fn_call_expr)]
                       )
    
    #t = closure_type.make_closure_type(typed_fn, get_types(closure_args))
    #return Closure(typed_fn, closure_args, t)
    outer_closure_args = tuple(outer_closure_args) + (x,)
  
    idx_closure_t = closure_type.make_closure_type(idx_fn, get_types(outer_closure_args))
    
    idx_closure = Closure(idx_fn, args = outer_closure_args, type = idx_closure_t)
    
    result_type, typed_fn, typed_combine = \
      specialize_IndexReduce(idx_closure, combine, n_indices, init)
    if not self.is_none(init):
      init = self.cast(init, typed_combine.return_type)
    return syntax.IndexReduce(shape = shape, 
                              fn = make_typed_closure(idx_closure, typed_fn),
                              combine = make_typed_closure(combine, typed_combine),
                              init = init,   
                              type = result_type)
    
    
    
  def transform_Reduce(self, expr):
    new_args = self.transform_args(expr.args, flat = True)
    arg_types = get_types(new_args)
    axis = self.transform_if_expr(expr.axis)

    map_fn = self.transform_expr(expr.fn if expr.fn else untyped_identity_function) 
    combine_fn = self.transform_expr(expr.combine)
    
    init = self.transform_expr(expr.init) if expr.init else None
    
    # if there aren't any arrays, just treat this as a function call
    if all(isinstance(t, ScalarT) for t in arg_types):
      return self.invoke(map_fn, new_args)
    
    init_type = init.type if init else None
    
    if self.is_none(axis):
      if adverb_helpers.max_rank(arg_types) > 1:
        assert len(new_args) == 1, \
          "Can't handle multiple reduction inputs and flattening from axis=None"
        #x = new_args[0]
        #return self.flatten_Reduce(map_fn, combine_fn, x, init)
        #new_args = [self.ravel(new_args[0])]
        #arg_types = get_types(new_args)
        
        # Expect that the arguments will get raveled before 
        # the adverb gets evaluated 
        axis = self.none
        arg_types = [array_type.lower_rank(t, t.rank - 1) 
                     for t in arg_types
                     if t.rank > 1]
      else:
        axis = self.int(0)                        
    
    result_type, typed_map_fn, typed_combine_fn = \
        specialize_Reduce(map_fn.type,
                          combine_fn.type,
                          arg_types, 
                          init_type)
    typed_map_closure = make_typed_closure (map_fn, typed_map_fn)
    typed_combine_closure = make_typed_closure(combine_fn, typed_combine_fn)
    
    
    if init_type and init_type != result_type and \
       array_type.rank(init_type) < array_type.rank(result_type):
      assert len(new_args) == 1
      #assert is_zero(axis), "Unexpected axis %s : %s" % (axis, axis.type)
      arg = new_args[0]
      first_elt = syntax.Index(arg, zero_i64, 
                                  type = arg.type.index_type(zero_i64))
      first_combine = specialize(combine_fn, (init_type, first_elt.type))
      first_combine_closure = make_typed_closure(combine_fn, first_combine)
      init = syntax.Call(first_combine_closure, (init, first_elt), 
                                 type = first_combine.return_type)
      slice_rest = syntax.Slice(start = one_i64, stop = none, step = one_i64, 
                                   type = array_type.SliceT(Int64, NoneType, Int64))
      rest = syntax.Index(arg, slice_rest, 
                             type = arg.type.index_type(slice_rest))
      new_args = (rest,)  
    
    return syntax.Reduce(fn = typed_map_closure,
                         combine = typed_combine_closure,
                         args = new_args,
                         axis = axis,
                         type = result_type,
                         init = init)

  def transform_Scan(self, expr):
    map_fn = self.transform_expr(expr.fn if expr.fn else untyped_identity_function)
    combine_fn = self.transform_expr(expr.combine)
    emit_fn = self.transform_expr(expr.emit)
    new_args = self.transform_args(expr.args, flat = True)
    arg_types = get_types(new_args)
    init = self.transform_expr(expr.init) if expr.init else None
    init_type = get_type(init) if init else None
    result_type, typed_map_fn, typed_combine_fn, typed_emit_fn = \
        specialize_Scan(map_fn.type, combine_fn.type, emit_fn.type,
                        arg_types, init_type)
    map_fn.fn = typed_map_fn
    combine_fn.fn = typed_combine_fn
    emit_fn.fn = typed_emit_fn
    axis = self.transform_if_expr(expr.axis)
    if axis is None or self.is_none(axis):
      assert adverb_helpers.max_rank(arg_types) == 1
      axis = syntax_helpers.zero_i64
    return syntax.Scan(fn = make_typed_closure(map_fn, typed_map_fn),
                       combine = make_typed_closure(combine_fn,
                                                    typed_combine_fn),
                       emit = make_typed_closure(emit_fn, typed_emit_fn),
                       args = new_args,
                       axis = axis,
                       type = result_type,
                       init = init)

  def transform_AllPairs(self, expr):
    closure = self.transform_expr(expr.fn)
    new_args = self.transform_args (expr.args, flat = True)
    arg_types = get_types(new_args)
    assert len(arg_types) == 2
    xt,yt = arg_types
    result_type, typed_fn = specialize_AllPairs(closure.type, xt, yt)
    axis = self.transform_if_expr(expr.axis)
    if axis is None or self.is_none(axis):
      axis = zero_i64
    return syntax.AllPairs(fn = make_typed_closure(closure, typed_fn),
                           args = new_args,
                           axis = axis,
                           type = result_type)
  

  
  def infer_phi(self, result_var, val):
    """
    Don't actually rewrite the phi node, just add any necessary types to the
    type environment
    """

    new_val = self.transform_expr(val)
    new_type = new_val.type
    old_type = self.type_env.get(result_var, Unknown)
    new_result_var = self.var_map.lookup(result_var)
    self.type_env[new_result_var]  = old_type.combine(new_type)

  def infer_phi_nodes(self, nodes, direction):
    for (var, values) in nodes.iteritems():
      self.infer_phi(var, direction(values))

  def infer_left_flow(self, nodes):
    return self.infer_phi_nodes(nodes, lambda (x, _): x)

  def infer_right_flow(self, nodes):
    return self.infer_phi_nodes(nodes, lambda (_, x): x)

  
  def transform_phi_node(self, result_var, (left_val, right_val)):
    """
    Rewrite the phi node by rewriting the values from either branch, renaming
    the result variable, recording its new type, and returning the new name
    paired with the annotated branch values
    """

    new_left = self.transform_expr(left_val)
    new_right = self.transform_expr(right_val)
    old_type = self.type_env.get(result_var, Unknown)
    new_type = old_type.combine(new_left.type).combine(new_right.type)
    new_var = self.var_map.lookup(result_var)
    self.type_env[new_var] = new_type
    return (new_var, (new_left, new_right))

  def transform_phi_nodes(self, nodes):
    new_nodes = {}
    for old_k, (old_left, old_right) in nodes.iteritems():
      new_name, (left, right) = self.transform_phi_node(old_k, (old_left, old_right))
      new_nodes[new_name] = (left, right)
    return new_nodes

  def annotate_lhs(self, lhs, rhs_type):

    lhs_class = lhs.__class__
    if lhs_class is syntax.Tuple:
      if rhs_type.__class__ is TupleT:
        assert len(lhs.elts) == len(rhs_type.elt_types)
        new_elts = [self.annotate_lhs(elt, elt_type) 
                    for (elt, elt_type) in zip(lhs.elts, rhs_type.elt_types)]
      else:
        assert rhs_type.__class__ is ArrayT, \
            "Unexpected right hand side type %s for %s" % (rhs_type, lhs)
        elt_type = array_type.lower_rank(rhs_type, 1)
        new_elts = [self.annotate_lhs(elt, elt_type) for elt in lhs.elts]
      tuple_t = tuple_type.make_tuple_type(get_types(new_elts))
      return syntax.Tuple(new_elts, type = tuple_t)
    elif lhs_class is syntax.Index:
      new_arr = self.transform_expr(lhs.value)
      new_idx = self.transform_expr(lhs.index)
      assert isinstance(new_arr.type, ArrayT), \
          "Expected array, got %s" % new_arr.type
      elt_t = new_arr.type.index_type(new_idx.type)
      return syntax.Index(new_arr, new_idx, type = elt_t)
    elif lhs_class is syntax.Attribute:
      name = lhs.name
      struct = self.transform_expr(lhs.value)
      struct_t = struct.type
      assert isinstance(struct_t, StructT), \
          "Can't access fields on value %s of type %s" % \
          (struct, struct_t)
      field_t = struct_t.field_type(name)
      return syntax.Attribute(struct, name, field_t)
    else:
      assert lhs_class is Var, "Unexpected LHS: %s" % (lhs,)
      new_name = self.var_map.lookup(lhs.name)
      old_type = self.type_env.get(new_name, Unknown)
      new_type = old_type.combine(rhs_type)
      self.type_env[new_name] = new_type
      return Var(new_name, type = new_type)

  def transform_Assign(self, stmt):
    rhs = self.transform_expr(stmt.rhs)
    lhs = self.annotate_lhs(stmt.lhs, rhs.type)
    return syntax.Assign(lhs, rhs)
  
  
  
  def transform_If(self, stmt):
    cond = self.transform_expr(stmt.cond) 

    assert isinstance(cond.type, ScalarT), \
        "Condition %s has type %s but must be convertible to bool" % (cond, cond.type)
    # it would be cleaner to not have anything resembling an optimization 
    # inter-mixed with the type inference, but I'm not sure how else to 
    # support 'if x is None:...'
    if is_true(cond):
      self.blocks.top().extend(self.transform_block(stmt.true))
      for (name, (left,_)) in stmt.merge.iteritems():
        typed_left = self.transform_expr(left)
        typed_var = self.annotate_lhs(Var(name), typed_left.type) 
        self.assign(typed_var, typed_left)
      return
    
    if is_false(cond):
      self.blocks.top().extend(self.transform_block(stmt.false))
      for (name, (_,right)) in stmt.merge.iteritems():
        typed_right = self.transform_expr(right)
        typed_var = self.annotate_lhs(Var(name), typed_right.type)
        self.assign(typed_var, typed_right)
      return
    true = self.transform_block(stmt.true)
    false = self.transform_block(stmt.false) 
    merge = self.transform_phi_nodes(stmt.merge)
    return syntax.If(cond, true, false, merge)

  def transform_Return(self, stmt):
    ret_val = self.transform_expr(stmt.value)
    curr_return_type = self.type_env["$return"]
    self.type_env["$return"] = curr_return_type.combine(ret_val.type)
    return syntax.Return(ret_val)

  def transform_While(self, stmt):
    self.infer_left_flow(stmt.merge)
    cond = self.transform_expr(stmt.cond)
    body = self.transform_block(stmt.body)
    merge = self.transform_phi_nodes(stmt.merge)
    return syntax.While(cond, body, merge)

  def transform_ForLoop(self, stmt):
    self.infer_left_flow(stmt.merge)
    start = self.transform_expr(stmt.start)
    stop = self.transform_expr(stmt.stop)
    step = self.transform_expr(stmt.step)
    lhs_t = start.type.combine(stop.type).combine(step.type)
    var = self.annotate_lhs(stmt.var, lhs_t)
    body = self.transform_block(stmt.body)
    merge = self.transform_phi_nodes(stmt.merge)
    return syntax.ForLoop(var, start, stop, step, body, merge)



def infer_types(untyped_fn, types):
  """
  Given an untyped function and input types, propagate the types through the
  body, annotating the AST with type annotations.

  NOTE: The AST won't be in a correct state until a rewrite pass back-propagates
  inferred types throughout the program and inserts adverbs for scalar operators
  applied to arrays
  """

  
  var_map = VarMap()
  typed_args = untyped_fn.args.transform(rename_fn = var_map.rename)
  if untyped_fn.args.starargs:
    assert typed_args.starargs

  unbound_keywords = []
  def keyword_fn(local_name, value):
    unbound_keywords.append(local_name)
    
    return type_conv.typeof(value)

  try: 
    tenv = typed_args.bind(types,
                           keyword_fn = keyword_fn,
                           starargs_fn = tuple_type.make_tuple_type)
  except  MissingArgsError as e:
    e.fn_name = untyped_fn.name 
    raise e 
    
  except: 
    print "Error while calling %s with types %s" % (untyped_fn, types)
    raise
  # keep track of the return
  tenv['$return'] = Unknown
  annotator = Annotator(tenv, var_map)
  body = annotator.transform_block(untyped_fn.body)
  arg_names = [local_name for local_name
               in
               typed_args.nonlocals + tuple(typed_args.positional)
               if local_name not in unbound_keywords]
  if len(unbound_keywords) > 0:
    default_assignments = []
    for local_name in unbound_keywords:
      t = tenv[local_name]
      python_value = typed_args.defaults[local_name]
      var = Var(local_name, type = t)
      if isinstance(python_value, tuple):
        parakeet_elts = []
        for (elt_value, elt_type) in zip(python_value, t.elt_types):
          parakeet_elt = syntax.Const(elt_value, elt_type)
          parakeet_elts.append(parakeet_elt)
        typed_val = syntax.Tuple(tuple(parakeet_elts), type = t)
      else:
        typed_val = syntax.Const(python_value, t) #mk_default_const(python_value, t)

      stmt = syntax.Assign(var, typed_val)
      default_assignments.append(stmt)
    body = default_assignments + body

  input_types = tuple([tenv[arg_name] for arg_name in arg_names])

  # starargs are all passed individually and then packaged up
  # into a tuple on the first line of the function
  if typed_args.starargs:
    local_starargs_name = typed_args.starargs

    starargs_t = tenv[local_starargs_name]
    assert starargs_t.__class__ is TupleT, \
        "Unexpected starargs type %s" % starargs_t
    extra_arg_vars = []
    for (i, elt_t) in enumerate(starargs_t.elt_types):
      arg_name = "%s_elt%d" % (names.original(local_starargs_name), i)
      tenv[arg_name] = elt_t
      arg_var = Var(name = arg_name, type = elt_t)
      arg_names.append(arg_name)
      extra_arg_vars.append(arg_var)
    input_types = input_types + starargs_t.elt_types
    tuple_lhs = Var(name = local_starargs_name, type = starargs_t)
    tuple_rhs = syntax.Tuple(elts = extra_arg_vars, type = starargs_t)
    stmt = syntax.Assign(tuple_lhs, tuple_rhs)
    body = [stmt] + body

  return_type = tenv["$return"]
  # if nothing ever gets returned, then set the return type to None
  if isinstance(return_type,  UnknownT):
    body.append(syntax.Return(syntax_helpers.none))
    tenv["$return"] = NoneType
    return_type = NoneType

  return syntax.TypedFn(
    name = names.refresh(untyped_fn.name),
    body = body,
    arg_names = arg_names,
    input_types = input_types,
    return_type = return_type,
    type_env = tenv)

def _specialize(fn, arg_types, return_type = None):
  """
  Do the actual work of type specialization, whereas the wrapper 'specialize'
  pulls out untyped functions from closures, wraps argument lists in ActualArgs
  objects and performs memoization
  """

  if isinstance(fn, syntax.TypedFn):
    return fn
  typed_fundef = infer_types(fn, arg_types)
  from rewrite_typed import rewrite_typed
  coerced_fundef = rewrite_typed(typed_fundef, return_type)
  import simplify
  normalized = simplify.Simplify().apply(coerced_fundef)
  return normalized

def _get_fundef(fn):
  if isinstance(fn, (syntax.Fn, syntax.TypedFn)):
    return fn
  else:
    assert isinstance(fn, str), \
        "Unexpected function %s : %s"  % (fn, fn.type)
    return syntax.Fn.registry[fn]

def _get_closure_type(fn):
  assert isinstance(fn, (Fn, TypedFn, ClosureT, Closure, Var)), \
    "Expected function, got %s" % fn
    
  if fn.__class__ is closure_type.ClosureT:
    return fn
  elif isinstance(fn, Closure):
    return fn.type
  elif isinstance(fn, Var):
    assert isinstance(fn.type, closure_type.ClosureT)
    return fn.type
  else:
    fundef = _get_fundef(fn)
    return closure_type.make_closure_type(fundef, [])

def specialize(fn, arg_types, return_type = None):
  if config.print_before_specialization:
    if return_type:
      print "==== Specializing", fn, "for input types", arg_types, "and return type", return_type
    else:  
      print "=== Specializing", fn, "for types", arg_types 
  if isinstance(fn, syntax.TypedFn):
    assert len(fn.input_types) == len(arg_types)
    assert all(t1 == t2 for t1,t2 in zip(fn.input_types, arg_types))
    if return_type is not None:
      assert fn.return_type == return_type 
    return fn
  
  if isinstance(arg_types, (list, tuple)):
    arg_types = ActualArgs(arg_types)
  closure_t = _get_closure_type(fn)
  key = arg_types, return_type
  if key in closure_t.specializations:
    return closure_t.specializations[key]

  full_arg_types = arg_types.prepend_positional(closure_t.arg_types)
  fundef = _get_fundef(closure_t.fn)
  typed =  _specialize(fundef, full_arg_types, return_type)
  closure_t.specializations[key] = typed

  if config.print_specialized_function:
    if return_type:
      print "=== Specialized %s for input types %s and return type %s ==="  % \
          (fundef.name, full_arg_types, return_type)
    else:
      print "=== Specialized %s for input types %s ==="  % \
          (fundef.name, full_arg_types)
    
    print
    print repr(typed)
    print

  return typed

def infer_return_type(untyped, arg_types):
  """
  Given a function definition and some input types, gives back the return type
  and implicitly generates a specialized version of the function.
  """

  typed = specialize(untyped, arg_types)
  return typed.return_type

def specialize_IndexMap(fn, n_indices):
  idx_type = make_tuple_type( (Int64,) * n_indices) if n_indices > 1 else Int64
  typed_fn = specialize(fn, (idx_type,))
  result_type = array_type.increase_rank(typed_fn.return_type, n_indices)
  return result_type, typed_fn

def specialize_IndexReduce(fn, combine, n_indices, init = None):
  idx_type = make_tuple_type( (Int64,) * n_indices) if n_indices > 1 else Int64
  if init is None or isinstance(init.type, NoneT):
    typed_fn = specialize(fn, (idx_type,))
  else:
    typed_fn = specialize(fn, (idx_type,), return_type = init.type)
  elt_type = typed_fn.return_type
  typed_combine = specialize(combine, (elt_type, elt_type))
  return elt_type, typed_fn, typed_combine 
      
def specialize_Map(map_fn, array_types):
  elt_types = array_type.lower_ranks(array_types, 1)
  typed_map_fn = specialize(map_fn, elt_types)
  elt_result_t = typed_map_fn.return_type
  result_t = array_type.increase_rank(elt_result_t, 1)
  return result_t, typed_map_fn

def infer_Map(map_fn, array_types):
  t, _ = specialize_Map(map_fn, array_types)
  return t

def specialize_Reduce(map_fn, combine_fn, array_types, init_type = None):
  _, typed_map_fn = specialize_Map(map_fn, array_types)
  elt_type = typed_map_fn.return_type
  if init_type is None or isinstance(init_type, NoneT):
    acc_type = elt_type
  else:
    acc_type = init_type

  typed_combine_fn = specialize(combine_fn, [acc_type, elt_type])
  new_acc_type = typed_combine_fn.return_type
  if new_acc_type != acc_type:
    typed_combine_fn = specialize(combine_fn, [new_acc_type, elt_type])
    new_acc_type = typed_combine_fn.return_type
  #assert new_acc_type == acc_type, \
  #  "Expected accumulator types %s but encountered %s" % (acc_type, new_acc_type)
  return new_acc_type, typed_map_fn, typed_combine_fn

def infer_Reduce(map_fn, combine_fn, array_types, init_type = None):
  t, _, _ = specialize_Reduce(map_fn, combine_fn, array_types, init_type)
  return t

def specialize_Scan(map_fn, combine_fn, emit_fn, array_types, init_type = None):
  acc_type, typed_map_fn, typed_combine_fn = \
      specialize_Reduce(map_fn, combine_fn, array_types, init_type)
  typed_emit_fn = specialize(emit_fn, [acc_type])
  result_type = array_type.increase_rank(typed_emit_fn.return_type, 1)
  return result_type, typed_map_fn, typed_combine_fn, typed_emit_fn

def infer_Scan(map_fn, combine_fn, emit_fn, array_types, init_type = None):
  t, _, _, _ = specialize_Scan(map_fn, combine_fn, emit_fn,
                               array_types, init_type)
  return t

def specialize_AllPairs(fn, xtype, ytype):
  x_elt_t = array_type.lower_rank(xtype, 1)
  y_elt_t = array_type.lower_rank(ytype, 1)
  typed_map_fn = specialize(fn, [x_elt_t, y_elt_t])
  elt_result_t = typed_map_fn.return_type
  result_t = array_type.increase_rank(elt_result_t, 2)
  return result_t, typed_map_fn

def infer_AllPairs(fn, xtype, ytype):
  t, _ = specialize_AllPairs(fn, xtype, ytype)
  return t




import args
import array_type
import core_types
import names
import syntax
import syntax_helpers

def max_rank(arg_types):
  """
  Given a list of types, find the maximum rank of the list and also check that
  all other types have either the same rank or are scalars
  """

  curr_max = 0
  for t in arg_types:
    if isinstance(t, array_type.ArrayT):
      assert curr_max == 0 or curr_max == t.rank,  \
       "Adverb can't accept inputs of rank %d and %d" % (curr_max, t.rank)
      curr_max = t.rank
  return curr_max

def max_rank_arg(args):
  """Given a list of arguments, return one which has the maximum rank"""

  r = max_rank(syntax_helpers.get_types(args))
  for arg in args:
    if arg.type.rank == r:
      return arg

def num_outer_axes(arg_types, axis):
  """
  Helper for adverb type inference to figure out how many axes it will loop over
  -- either 1 particular one or all of them when axis is None.
  """

  axis = syntax_helpers.unwrap_constant(axis)
  if isinstance(arg_types, core_types.Type):
    max_arg_rank = arg_types.rank
  else:
    max_arg_rank = max_rank(arg_types)
  return 1 if (max_arg_rank > 0 and axis is not None) else max_arg_rank

_nested_map_cache = {}
def nested_maps(inner_fn, depth, arg_names):
  if depth <= 0:
    return inner_fn

  key = inner_fn.name, depth, tuple(arg_names)
  if key in _nested_map_cache:
    return _nested_map_cache[key]
  args_obj = args.FormalArgs()
  arg_vars = []
  for var_name in arg_names:
    local_name = names.refresh(var_name)
    args_obj.add_positional(local_name)
    arg_vars.append(syntax.Var(local_name))

  name = names.fresh(inner_fn.name + "_broadcast%d" % depth)
  nested_fn = nested_maps(inner_fn, depth - 1, arg_names)

  map_expr = syntax.Map(fn = nested_fn, 
                        axis = syntax_helpers.zero_i64, 
                        args = arg_vars)
  fn = syntax.Fn(
    name = name,
    args = args_obj,
    body = [syntax.Return(map_expr)]
  )

  _nested_map_cache[key] = fn
  return fn





from parakeet import *
from prims import * 
from lib import * 

modules = [
  "adverb_helpers",
  "adverb_semantics",
  "adverb_wrapper",
  "adverbs",
  "args",
  "array_type",
  "ast_conversion",
  "c_function",
  "clone_function",
  "closure_type",
  "codegen",
  "collect_vars",
  "common",
  "config",
  "copy_elimination",
  "core_types",
  "dead_code_elim",
  "dtypes",
  "flow_analysis",
  "function_registry",
  "fusion",
  "inline",
  "interp",
  "lib_simple",
  "licm",
  "llvm_backend",
  "llvm_context",
  "llvm_convert",
  "llvm_helpers",
  "llvm_prims",
  "llvm_types",
  "lower_adverbs",
  "lower_indexing",
  "lower_structs",
  "lower_tiled_adverbs",
  "lowering",
  "mapify_allpairs",
  "memory",
  "mutability_analysis",
  "names",
  "nested_blocks",
  "node",
  "optimize",
  "prims",
  "python_ref",
  "rewrite_typed",
  "run_function",
  "scoped_dict",
  "scoped_set",
  "shape_codegen",
  "shape_eval",
  "shape_from_type",
  "shape_inference",
  "shape",
  "simplify",
  "subst",
  "syntax_helpers",
  "syntax_visitor",
  "syntax",
  "testing_helpers",
  "tile_adverbs",
  "transform",
  "traversal",
  "tuple_type",
  "type_conv",
  "type_conv_decls",
  "type_inference",
  "use_analysis",
  "verify"
]
